#========================================================================================#
"""
	IdealGas

Extending SimpleParticles to conserve kinetic energy and momentum.

Author: Niall Palfreyman, 20/01/2023.
"""
module IdealGas

using Agents, LinearAlgebra, GLMakie, InteractiveDynamics

#-----------------------------------------------------------------------------------------
# Module data:

"""
	Particle

The populating agents in the IdealGas model.
"""
@agent Particle ContinuousAgent{2} begin
	mass::Float64					# Particle's mass
	speed::Float64					# Particle's speed
	radius::Float64					# Particle's radius
	prev_partner::Int				# Previous collision partner id
end

"Standard value that is definitely NOT a valid agent ID"
const non_id = -1

#-----------------------------------------------------------------------------------------
# Module methods:

"""
	idealgas( kwargs)

Create and initialise the IdealGas model.
"""
function idealgas(;
    n_particles = 50,				# Number of Particles in box
	masses = [1.0],					# Possible masses of Particles
    init_speed = 1.0,				# Initial speed of Particles in box
	radius = 1,						# Radius of Particles in the box
    extent = (100, 40),				# Extent of Particles space
)
    space = ContinuousSpace(extent; spacing = radius/1.5)
    box = ABM( Particle, space; scheduler = Schedulers.Randomly())

    for _ in 1:n_particles
        vel = Tuple( 2rand(2).-1)
		vel = vel ./ norm(vel)		# ALWAYS maintain normalised state of vel!
        add_agent!( box, vel, rand(masses), init_speed, radius, non_id)
    end

    return box
end

#-----------------------------------------------------------------------------------------
"""
	agent_step!( me, box)

This is the heart of the IdealGas model: It calculates how Particles collide with each other,
while conserving momentum and kinetic energy.
"""
function agent_step!(me::Particle, box::ABM)
	her = random_nearby_agent( me, box, 2*me.radius)	# Grab nearby particle
	if her === nothing
		# No new partners - forget previous collision partner:
		me.prev_partner = non_id
	elseif her.id < me.id && her.id != me.prev_partner
		# New collision partner has not already been handled and is not my previous partner:
		me.prev_partner = her.id							# Update previous partners to avoid
		her.prev_partner = me.id							# repetitive juddering collisions.
		cntct = (x->[cos(x),sin(x)])(2rand()pi)				# Unit vector to contact point with partner
		Rctct = [cntct[1] cntct[2]; -cntct[2] cntct[1]]		# Rotation into contact directn coords
		Rback = [cntct[1] -cntct[2]; cntct[2] cntct[1]]		# Inverse rotation back to world coords

		# Rotate velocities into coordinates directed ALONG and PERPendicular to contact direction:
		myAlongVel, myPerpVel = me.speed * Rctct * collect(me.vel)					# My velocity
		herAlongVel, herPerpVel = her.speed * Rctct * collect(her.vel)				# Her velocity
		cmAlongVel = (me.mass*myAlongVel + her.mass*herAlongVel)/(me.mass+her.mass)	# C of M velocity

		# Calculate collision effects along contact direction (perp direction is unaffected):
		myAlongVel = 2cmAlongVel - myAlongVel
		herAlongVel = 2cmAlongVel - herAlongVel

		# Rotate collision effects on both me and her back into world coordinates:
		me.speed = hypot(myAlongVel,myPerpVel)
		if me.speed != 0.0
			me.vel = Tuple(Rback*[myAlongVel,myPerpVel])
			me.vel = me.vel ./ norm(me.vel)
		end
		her.speed = hypot(herAlongVel,herPerpVel)
		if her.speed != 0.0
			her.vel = Tuple(Rback*[herAlongVel,myPerpVel])
			her.vel = her.vel ./ norm(her.vel)
		end
	end

	move_agent!( me, box, me.speed)							# Advance me with my current speed
end

#-----------------------------------------------------------------------------------------
"""
	momentum( particle)

Return the momentum of this particle.
"""
function momentum(particle)
	particle.mass * particle.speed * collect(particle.vel)
end

#-----------------------------------------------------------------------------------------
"""
	kinetic_energy( particle)

Return the kinetic energy of this particle.
"""
function kinetic_energy(particle)
	particle.mass * particle.speed^2 / 2
end

#-----------------------------------------------------------------------------------------
"""
	demo()

Run a simulation of the IdealGas model.
"""
function demo()
	box = idealgas()
	abmvideo(
		"IdealGas.mp4", box, agent_step!;
		framerate = 20, frames = 200,
		title = "Simple particles in an ideal gas",
		ac=:blue, as=20, am=:circle
	)

	agentdata, = run!( box, agent_step!, 500; adata=[(momentum,sum),(kinetic_energy,sum)])

	# Return momentum and energy statistics:
	agentdata
end

end	# of module IdealGas